"""LCM type definitions
This file automatically generated by lcm.
DO NOT MODIFY BY HAND!!!!
"""


from io import BytesIO
import struct

class lcmt_planar_manipuland_status(object):

    __slots__ = ["utime", "position", "theta", "velocity", "thetadot"]

    __typenames__ = ["int64_t", "double", "double", "double", "double"]

    __dimensions__ = [None, [2], None, [2], None]

    def __init__(self):
        self.utime = 0
        """
        The timestamp in microseconds
        LCM Type: int64_t
        """

        self.position = [ 0.0 for dim0 in range(2) ]
        """
        The translational position of the manipuland.
        LCM Type: double[2]
        """

        self.theta = 0.0
        """
        The angle of the manipuland.
        LCM Type: double
        """

        self.velocity = [ 0.0 for dim0 in range(2) ]
        """
        The translational velocity of the manipuland.
        LCM Type: double[2]
        """

        self.thetadot = 0.0
        """
        The angular velocity of the manipuland.
        LCM Type: double
        """


    def encode(self):
        buf = BytesIO()
        buf.write(lcmt_planar_manipuland_status._get_packed_fingerprint())
        self._encode_one(buf)
        return buf.getvalue()

    def _encode_one(self, buf):
        buf.write(struct.pack(">q", self.utime))
        buf.write(struct.pack('>2d', *self.position[:2]))
        buf.write(struct.pack(">d", self.theta))
        buf.write(struct.pack('>2d', *self.velocity[:2]))
        buf.write(struct.pack(">d", self.thetadot))

    @staticmethod
    def decode(data: bytes):
        if hasattr(data, 'read'):
            buf = data
        else:
            buf = BytesIO(data)
        if buf.read(8) != lcmt_planar_manipuland_status._get_packed_fingerprint():
            raise ValueError("Decode error")
        return lcmt_planar_manipuland_status._decode_one(buf)

    @staticmethod
    def _decode_one(buf):
        self = lcmt_planar_manipuland_status()
        self.utime = struct.unpack(">q", buf.read(8))[0]
        self.position = struct.unpack('>2d', buf.read(16))
        self.theta = struct.unpack(">d", buf.read(8))[0]
        self.velocity = struct.unpack('>2d', buf.read(16))
        self.thetadot = struct.unpack(">d", buf.read(8))[0]
        return self

    @staticmethod
    def _get_hash_recursive(parents):
        if lcmt_planar_manipuland_status in parents: return 0
        tmphash = (0xd2ae3532484f4461) & 0xffffffffffffffff
        tmphash  = (((tmphash<<1)&0xffffffffffffffff) + (tmphash>>63)) & 0xffffffffffffffff
        return tmphash
    _packed_fingerprint = None

    @staticmethod
    def _get_packed_fingerprint():
        if lcmt_planar_manipuland_status._packed_fingerprint is None:
            lcmt_planar_manipuland_status._packed_fingerprint = struct.pack(">Q", lcmt_planar_manipuland_status._get_hash_recursive([]))
        return lcmt_planar_manipuland_status._packed_fingerprint

    def get_hash(self):
        """Get the LCM hash of the struct"""
        return struct.unpack(">Q", lcmt_planar_manipuland_status._get_packed_fingerprint())[0]




"""
Train a policy for //bindings/pydrake/examples/envs:cart_pole
"""

import argparse

import gymnasium as gym
import stable_baselines3
from stable_baselines3.common.env_checker import check_env

from pydrake.common import configure_logging
from pydrake.examples.gym._bazel_cwd_helpers import bazel_chdir
from pydrake.geometry import StartMeshcat

_full_sb3_available = False
if "drake_internal" not in stable_baselines3.__version__:
    from stable_baselines3 import PPO
    from stable_baselines3.common.callbacks import EvalCallback
    from stable_baselines3.common.env_util import make_vec_env
    from stable_baselines3.common.vec_env import (
        DummyVecEnv,
        SubprocVecEnv,
        VecVideoRecorder,
    )
    import torch as th

    _full_sb3_available = True


def _run_training(config, args):
    env_name = config["env_name"]
    num_env = config["num_workers"]
    time_limit = config["env_time_limit"]
    log_dir = config["local_log_dir"]
    policy_type = config["policy_type"]
    total_timesteps = config["total_timesteps"]
    policy_kwargs = config["policy_kwargs"]
    eval_freq = config["model_save_freq"]
    obs_noise = config["observation_noise"]
    add_disturbances = config["disturbances"]

    if not args.train_single_env:
        env = make_vec_env(
            env_name,
            n_envs=num_env,
            seed=0,
            vec_env_cls=SubprocVecEnv,
            env_kwargs={
                "time_limit": time_limit,
                "obs_noise": obs_noise,
                "add_disturbances": add_disturbances,
            },
        )
    else:
        meshcat = StartMeshcat()
        env = gym.make(
            env_name,
            meshcat=meshcat,
            time_limit=time_limit,
            debug=args.debug,
            obs_noise=obs_noise,
        )
        check_env(env)
        input("Open meshcat (optional). Press Enter to continue...")

    if args.test:
        model = PPO(
            policy_type,
            env,
            n_steps=4,
            n_epochs=2,
            batch_size=8,
            policy_kwargs=policy_kwargs,
        )
    else:
        tensorboard_log = f"{log_dir}runs/test"
        model = PPO(
            policy_type,
            env,
            n_steps=int(2048 / num_env),
            n_epochs=10,
            # In SB3, this is the mini-batch size.
            # https://github.com/DLR-RM/stable-baselines3/blob/master/docs/modules/ppo.rst
            batch_size=64 * num_env,
            verbose=1,
            tensorboard_log=tensorboard_log,
            policy_kwargs=policy_kwargs,
        )
        print(
            "Open tensorboard (optional) via "
            f"`tensorboard --logdir {tensorboard_log}` "
            "in another terminal."
        )

    # Separate evaluation env.
    eval_env = gym.make(
        env_name,
        time_limit=time_limit,
        obs_noise=obs_noise,
        monitoring_camera=True,
    )
    eval_env = DummyVecEnv([lambda: eval_env])
    # Record a video every n evaluation rollouts.
    n = 1
    eval_env = VecVideoRecorder(
        eval_env,
        log_dir + "videos/test",
        record_video_trigger=lambda x: x % n == 0,
        video_length=100,
    )
    # Use deterministic actions for evaluation.
    eval_callback = EvalCallback(
        eval_env,
        best_model_save_path=log_dir + "eval_logs/test",
        log_path=log_dir + "eval_logs/test",
        eval_freq=eval_freq,
        deterministic=True,
        render=False,
    )

    model.learn(
        total_timesteps=total_timesteps,
        callback=eval_callback,
    )

    eval_env.close()


def _main():
    bazel_chdir()
    configure_logging()
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--test", action="store_true")
    parser.add_argument("--train_single_env", action="store_true")
    parser.add_argument("--debug", action="store_true")
    parser.add_argument(
        "--log_path",
        help="path to the logs directory.",
        default="./rl/tmp/DrakeCartPole/",
    )
    args = parser.parse_args()

    if not _full_sb3_available:
        print("stable_baselines3 found, but was drake internal")
        return 0 if args.test else 1

    if args.test:
        num_env = 2
    elif args.train_single_env:
        num_env = 1
    else:
        num_env = 10

    config = {
        "policy_type": "MlpPolicy",
        "total_timesteps": 5e5 if not args.test else 5,
        "env_name": "DrakeCartPole-v0",
        "num_workers": num_env,
        "env_time_limit": 7 if not args.test else 0.5,
        "local_log_dir": args.log_path,
        "model_save_freq": 1e4,
        "policy_kwargs": {
            "activation_fn": th.nn.ReLU,
            "net_arch": {"pi": [64, 64, 64], "vf": [64, 64, 64]},
        },
        "observation_noise": True,
        "disturbances": True,
    }

    _run_training(config, args)


gym.envs.register(
    id="DrakeCartPole-v0",
    entry_point=("pydrake.examples.gym.envs.cart_pole:DrakeCartPoleEnv"),
)

if __name__ == "__main__":
    _main()



"""LCM type definitions
This file automatically generated by lcm.
DO NOT MODIFY BY HAND!!!!
"""


from io import BytesIO
import struct

class lcmt_schunk_wsg_status(object):

    __slots__ = ["utime", "actual_position_mm", "actual_force", "actual_speed_mm_per_s"]

    __typenames__ = ["int64_t", "double", "double", "double"]

    __dimensions__ = [None, None, None, None]

    def __init__(self):
        self.utime = 0
        """ LCM Type: int64_t """
        self.actual_position_mm = 0.0
        """
        < The timestamp in microseconds.
        LCM Type: double
        """

        self.actual_force = 0.0
        """
        The combined force being applied by both gripper fingers in newtons.
        While some implementations may report a negative value for this field
        under some circumstances, it should always be interpreted as an unsigned
        quantity (absolute value).
        LCM Type: double
        """

        self.actual_speed_mm_per_s = 0.0
        """ LCM Type: double """

    def encode(self):
        buf = BytesIO()
        buf.write(lcmt_schunk_wsg_status._get_packed_fingerprint())
        self._encode_one(buf)
        return buf.getvalue()

    def _encode_one(self, buf):
        buf.write(struct.pack(">qddd", self.utime, self.actual_position_mm, self.actual_force, self.actual_speed_mm_per_s))

    @staticmethod
    def decode(data: bytes):
        if hasattr(data, 'read'):
            buf = data
        else:
            buf = BytesIO(data)
        if buf.read(8) != lcmt_schunk_wsg_status._get_packed_fingerprint():
            raise ValueError("Decode error")
        return lcmt_schunk_wsg_status._decode_one(buf)

    @staticmethod
    def _decode_one(buf):
        self = lcmt_schunk_wsg_status()
        self.utime, self.actual_position_mm, self.actual_force, self.actual_speed_mm_per_s = struct.unpack(">qddd", buf.read(32))
        return self

    @staticmethod
    def _get_hash_recursive(parents):
        if lcmt_schunk_wsg_status in parents: return 0
        tmphash = (0x7635771fe2d72309) & 0xffffffffffffffff
        tmphash  = (((tmphash<<1)&0xffffffffffffffff) + (tmphash>>63)) & 0xffffffffffffffff
        return tmphash
    _packed_fingerprint = None

    @staticmethod
    def _get_packed_fingerprint():
        if lcmt_schunk_wsg_status._packed_fingerprint is None:
            lcmt_schunk_wsg_status._packed_fingerprint = struct.pack(">Q", lcmt_schunk_wsg_status._get_hash_recursive([]))
        return lcmt_schunk_wsg_status._packed_fingerprint

    def get_hash(self):
        """Get the LCM hash of the struct"""
        return struct.unpack(">Q", lcmt_schunk_wsg_status._get_packed_fingerprint())[0]




"""LCM type definitions
This file automatically generated by lcm.
DO NOT MODIFY BY HAND!!!!
"""


from io import BytesIO
import struct

import drake

class lcmt_viewer_link_data(object):

    __slots__ = ["name", "robot_num", "num_geom", "geom"]

    __typenames__ = ["string", "int32_t", "int32_t", "drake.lcmt_viewer_geometry_data"]

    __dimensions__ = [None, None, None, ["num_geom"]]

    def __init__(self):
        self.name = ""
        """ LCM Type: string """
        self.robot_num = 0
        """ LCM Type: int32_t """
        self.num_geom = 0
        """ LCM Type: int32_t """
        self.geom = []
        """ LCM Type: drake.lcmt_viewer_geometry_data[num_geom] """

    def encode(self):
        buf = BytesIO()
        buf.write(lcmt_viewer_link_data._get_packed_fingerprint())
        self._encode_one(buf)
        return buf.getvalue()

    def _encode_one(self, buf):
        __name_encoded = self.name.encode('utf-8')
        buf.write(struct.pack('>I', len(__name_encoded)+1))
        buf.write(__name_encoded)
        buf.write(b"\0")
        buf.write(struct.pack(">ii", self.robot_num, self.num_geom))
        for i0 in range(self.num_geom):
            assert self.geom[i0]._get_packed_fingerprint() == drake.lcmt_viewer_geometry_data._get_packed_fingerprint()
            self.geom[i0]._encode_one(buf)

    @staticmethod
    def decode(data: bytes):
        if hasattr(data, 'read'):
            buf = data
        else:
            buf = BytesIO(data)
        if buf.read(8) != lcmt_viewer_link_data._get_packed_fingerprint():
            raise ValueError("Decode error")
        return lcmt_viewer_link_data._decode_one(buf)

    @staticmethod
    def _decode_one(buf):
        self = lcmt_viewer_link_data()
        __name_len = struct.unpack('>I', buf.read(4))[0]
        self.name = buf.read(__name_len)[:-1].decode('utf-8', 'replace')
        self.robot_num, self.num_geom = struct.unpack(">ii", buf.read(8))
        self.geom = []
        for i0 in range(self.num_geom):
            self.geom.append(drake.lcmt_viewer_geometry_data._decode_one(buf))
        return self

    @staticmethod
    def _get_hash_recursive(parents):
        if lcmt_viewer_link_data in parents: return 0
        newparents = parents + [lcmt_viewer_link_data]
        tmphash = (0x4b645ec7a5743a2a+ drake.lcmt_viewer_geometry_data._get_hash_recursive(newparents)) & 0xffffffffffffffff
        tmphash  = (((tmphash<<1)&0xffffffffffffffff) + (tmphash>>63)) & 0xffffffffffffffff
        return tmphash
    _packed_fingerprint = None

    @staticmethod
    def _get_packed_fingerprint():
        if lcmt_viewer_link_data._packed_fingerprint is None:
            lcmt_viewer_link_data._packed_fingerprint = struct.pack(">Q", lcmt_viewer_link_data._get_hash_recursive([]))
        return lcmt_viewer_link_data._packed_fingerprint

    def get_hash(self):
        """Get the LCM hash of the struct"""
        return struct.unpack(">Q", lcmt_viewer_link_data._get_packed_fingerprint())[0]




import flags
import numpy
import pydrake.math
import pydrake.multibody.plant
import pydrake.symbolic
import pydrake.systems.analysis
import pydrake.systems.framework
import pydrake.systems.primitives
import pydrake.trajectories
from _typeshed import Incomplete
from typing import Any, Callable, ClassVar, overload

class DynamicProgrammingOptions:
    class PeriodicBoundaryCondition:
        high: float
        low: float
        state_index: int
        def __init__(self, state_index: int, low: float, high: float) -> None: ...
    assume_non_continuous_states_are_fixed: bool
    convergence_tol: float
    discount_factor: float
    input_port_index: pydrake.systems.framework.InputPortSelection | pydrake.systems.framework.InputPortIndex
    periodic_boundary_conditions: list[DynamicProgrammingOptions.PeriodicBoundaryCondition]
    visualization_callback: Callable[[int, pydrake.math.BarycentricMesh, numpy.ndarray[numpy.float64[1, n]], numpy.ndarray[numpy.float64[m, n]]], None]
    def __init__(self) -> None: ...

class FiniteHorizonLinearQuadraticRegulatorOptions:
    N: numpy.ndarray[numpy.float64[m, n]] | None
    Qf: numpy.ndarray[numpy.float64[m, n]] | None
    input_port_index: pydrake.systems.framework.InputPortSelection | pydrake.systems.framework.InputPortIndex
    simulator_config: pydrake.systems.analysis.SimulatorConfig
    u0: pydrake.trajectories.Trajectory
    ud: Incomplete
    use_square_root_method: bool
    x0: pydrake.trajectories.Trajectory
    xd: Any
    def __init__(self) -> None: ...

class FiniteHorizonLinearQuadraticRegulatorResult:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def K(self) -> pydrake.trajectories.Trajectory: ...
    @property
    def S(self) -> pydrake.trajectories.Trajectory: ...
    @property
    def k0(self) -> pydrake.trajectories.Trajectory: ...
    @property
    def s0(self) -> pydrake.trajectories.Trajectory: ...
    @property
    def sx(self) -> pydrake.trajectories.Trajectory: ...
    @property
    def u0(self) -> pydrake.trajectories.Trajectory: ...
    @property
    def x0(self) -> pydrake.trajectories.Trajectory: ...

class InverseDynamics(pydrake.systems.framework.LeafSystem):
    class InverseDynamicsMode:
        __members__: ClassVar[dict] = ...  # read-only
        __entries: ClassVar[dict] = ...
        kGravityCompensation: ClassVar[InverseDynamics.InverseDynamicsMode] = ...
        kInverseDynamics: ClassVar[InverseDynamics.InverseDynamicsMode] = ...
        def __init__(self, value: int) -> None: ...
        def __eq__(self, other: object) -> bool: ...
        def __hash__(self) -> int: ...
        def __index__(self) -> int: ...
        def __int__(self) -> int: ...
        def __ne__(self, other: object) -> bool: ...
        @property
        def name(self) -> str: ...
        @property
        def value(self) -> int: ...
    kGravityCompensation: ClassVar[InverseDynamics.InverseDynamicsMode] = ...
    kInverseDynamics: ClassVar[InverseDynamics.InverseDynamicsMode] = ...
    def __init__(self, plant: pydrake.multibody.plant.MultibodyPlant, mode: InverseDynamics.InverseDynamicsMode = ..., plant_context: pydrake.systems.framework.Context = ...) -> None: ...
    def get_input_port_desired_acceleration(self) -> pydrake.systems.framework.InputPort: ...
    def get_input_port_estimated_state(self) -> pydrake.systems.framework.InputPort: ...
    def get_output_port_generalized_force(self) -> pydrake.systems.framework.OutputPort: ...
    def is_pure_gravity_compensation(self) -> bool: ...

class InverseDynamicsController(pydrake.systems.framework.Diagram):
    def __init__(self, robot: pydrake.multibody.plant.MultibodyPlant, kp: numpy.ndarray[numpy.float64[m, 1]], ki: numpy.ndarray[numpy.float64[m, 1]], kd: numpy.ndarray[numpy.float64[m, 1]], has_reference_acceleration: bool, plant_context: pydrake.systems.framework.Context = ...) -> None: ...
    def get_input_port_desired_acceleration(self) -> pydrake.systems.framework.InputPort: ...
    def get_input_port_desired_state(self) -> pydrake.systems.framework.InputPort: ...
    def get_input_port_estimated_state(self) -> pydrake.systems.framework.InputPort: ...
    def get_multibody_plant_for_control(self) -> pydrake.multibody.plant.MultibodyPlant: ...
    def get_output_port_control(self) -> pydrake.systems.framework.OutputPort: ...
    def get_output_port_generalized_force(self) -> pydrake.systems.framework.OutputPort: ...
    def set_integral_value(self, arg0: pydrake.systems.framework.Context, arg1: numpy.ndarray[numpy.float64[m, 1]]) -> None: ...

class JointStiffnessController(pydrake.systems.framework.LeafSystem):
    def __init__(self, plant: pydrake.multibody.plant.MultibodyPlant, kp: numpy.ndarray[numpy.float64[m, 1]], kd: numpy.ndarray[numpy.float64[m, 1]]) -> None: ...
    def get_input_port_desired_state(self) -> pydrake.systems.framework.InputPort: ...
    def get_input_port_estimated_state(self) -> pydrake.systems.framework.InputPort: ...
    def get_multibody_plant(self) -> pydrake.multibody.plant.MultibodyPlant: ...
    def get_output_port_actuation(self) -> pydrake.systems.framework.OutputPort: ...

class PeriodicBoundaryCondition:
    high: float
    low: float
    state_index: int
    def __init__(self, state_index: int, low: float, high: float) -> None: ...

class PidControlledSystem(pydrake.systems.framework.Diagram):
    @overload
    def __init__(self, plant: pydrake.systems.framework.System, kp: float, ki: float, kd: float, state_output_port_index: int = ..., plant_input_port_index: int = ...) -> None: ...
    @overload
    def __init__(self, plant: pydrake.systems.framework.System, kp: numpy.ndarray[numpy.float64[m, 1]], ki: numpy.ndarray[numpy.float64[m, 1]], kd: numpy.ndarray[numpy.float64[m, 1]], state_output_port_index: int = ..., plant_input_port_index: int = ...) -> None: ...
    @overload
    def __init__(self, plant: pydrake.systems.framework.System, feedback_selector: numpy.ndarray[numpy.float64[m, n]], kp: float, ki: float, kd: float, state_output_port_index: int = ..., plant_input_port_index: int = ...) -> None: ...
    def get_control_input_port(self) -> pydrake.systems.framework.InputPort: ...
    def get_state_input_port(self) -> pydrake.systems.framework.InputPort: ...
    def get_state_output_port(self) -> pydrake.systems.framework.OutputPort: ...

class PidController(pydrake.systems.framework.LeafSystem):
    @overload
    def __init__(self, kp: numpy.ndarray[numpy.float64[m, 1]], ki: numpy.ndarray[numpy.float64[m, 1]], kd: numpy.ndarray[numpy.float64[m, 1]]) -> None: ...
    @overload
    def __init__(self, state_projection: numpy.ndarray[numpy.float64[m, n]], kp: numpy.ndarray[numpy.float64[m, 1]], ki: numpy.ndarray[numpy.float64[m, 1]], kd: numpy.ndarray[numpy.float64[m, 1]]) -> None: ...
    @overload
    def __init__(self, state_projection: numpy.ndarray[numpy.float64[m, n]], output_projection: numpy.ndarray[numpy.float64[m, n]], kp: numpy.ndarray[numpy.float64[m, 1]], ki: numpy.ndarray[numpy.float64[m, 1]], kd: numpy.ndarray[numpy.float64[m, 1]]) -> None: ...
    def get_Kd_vector(self) -> numpy.ndarray[numpy.float64[m, 1]]: ...
    def get_Ki_vector(self) -> numpy.ndarray[numpy.float64[m, 1]]: ...
    def get_Kp_vector(self) -> numpy.ndarray[numpy.float64[m, 1]]: ...
    def get_input_port_desired_state(self) -> pydrake.systems.framework.InputPort: ...
    def get_input_port_estimated_state(self) -> pydrake.systems.framework.InputPort: ...
    def get_output_port_control(self) -> pydrake.systems.framework.OutputPort: ...

def DiscreteTimeLinearQuadraticRegulator(A: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], B: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], Q: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], R: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], N: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous] = ...) -> tuple[numpy.ndarray[numpy.float64[m, n]], numpy.ndarray[numpy.float64[m, n]]]: ...
def FiniteHorizonLinearQuadraticRegulator(system: pydrake.systems.framework.System, context: pydrake.systems.framework.Context, t0: float, tf: float, Q: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], R: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], options: FiniteHorizonLinearQuadraticRegulatorOptions = ...) -> FiniteHorizonLinearQuadraticRegulatorResult: ...
def FittedValueIteration(arg0: pydrake.systems.analysis.Simulator, arg1: Callable[[pydrake.systems.framework.Context], float], arg2: list[set[float]], arg3: list[set[float]], arg4: float, arg5: DynamicProgrammingOptions) -> tuple[pydrake.systems.primitives.BarycentricMeshSystem, numpy.ndarray[numpy.float64[1, n]]]: ...
def LinearProgrammingApproximateDynamicProgramming(arg0: pydrake.systems.analysis.Simulator, arg1: Callable[[pydrake.systems.framework.Context], float], arg2: Callable[[numpy.ndarray[numpy.float64[m, 1]], numpy.ndarray[object[m, 1]]], pydrake.symbolic.Expression], arg3: int, arg4: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], arg5: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], arg6: float, arg7: DynamicProgrammingOptions) -> numpy.ndarray[numpy.float64[m, 1]]: ...
def LinearQuadraticRegulator(A: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], B: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], Q: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], R: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], N: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous] = ..., F: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous] = ...) -> tuple[numpy.ndarray[numpy.float64[m, n]], numpy.ndarray[numpy.float64[m, n]]]: ...
def MakeFiniteHorizonLinearQuadraticRegulator(system: pydrake.systems.framework.System, context: pydrake.systems.framework.Context, t0: float, tf: float, Q: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], R: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], options: FiniteHorizonLinearQuadraticRegulatorOptions = ...) -> pydrake.systems.framework.System: ...



"""LCM type definitions
This file automatically generated by lcm.
DO NOT MODIFY BY HAND!!!!
"""


from io import BytesIO
import struct

class lcmt_planar_gripper_finger_command(object):

    __slots__ = ["joint_position", "joint_velocity", "joint_torque"]

    __typenames__ = ["double", "double", "double"]

    __dimensions__ = [[2], [2], [2]]

    def __init__(self):
        self.joint_position = [ 0.0 for dim0 in range(2) ]
        """
        Only used when the planar gripper is in joint position control mode.
        Otherwise, ignored.
        LCM Type: double[2]
        """

        self.joint_velocity = [ 0.0 for dim0 in range(2) ]
        """ LCM Type: double[2] """
        self.joint_torque = [ 0.0 for dim0 in range(2) ]
        """
        Only used when the planar gripper is in joint torque control mode.
        Otherwise, ignored.
        LCM Type: double[2]
        """


    def encode(self):
        buf = BytesIO()
        buf.write(lcmt_planar_gripper_finger_command._get_packed_fingerprint())
        self._encode_one(buf)
        return buf.getvalue()

    def _encode_one(self, buf):
        buf.write(struct.pack('>2d', *self.joint_position[:2]))
        buf.write(struct.pack('>2d', *self.joint_velocity[:2]))
        buf.write(struct.pack('>2d', *self.joint_torque[:2]))

    @staticmethod
    def decode(data: bytes):
        if hasattr(data, 'read'):
            buf = data
        else:
            buf = BytesIO(data)
        if buf.read(8) != lcmt_planar_gripper_finger_command._get_packed_fingerprint():
            raise ValueError("Decode error")
        return lcmt_planar_gripper_finger_command._decode_one(buf)

    @staticmethod
    def _decode_one(buf):
        self = lcmt_planar_gripper_finger_command()
        self.joint_position = struct.unpack('>2d', buf.read(16))
        self.joint_velocity = struct.unpack('>2d', buf.read(16))
        self.joint_torque = struct.unpack('>2d', buf.read(16))
        return self

    @staticmethod
    def _get_hash_recursive(parents):
        if lcmt_planar_gripper_finger_command in parents: return 0
        tmphash = (0x1c9f9975f0aea67b) & 0xffffffffffffffff
        tmphash  = (((tmphash<<1)&0xffffffffffffffff) + (tmphash>>63)) & 0xffffffffffffffff
        return tmphash
    _packed_fingerprint = None

    @staticmethod
    def _get_packed_fingerprint():
        if lcmt_planar_gripper_finger_command._packed_fingerprint is None:
            lcmt_planar_gripper_finger_command._packed_fingerprint = struct.pack(">Q", lcmt_planar_gripper_finger_command._get_hash_recursive([]))
        return lcmt_planar_gripper_finger_command._packed_fingerprint

    def get_hash(self):
        """Get the LCM hash of the struct"""
        return struct.unpack(">Q", lcmt_planar_gripper_finger_command._get_packed_fingerprint())[0]




"""LCM type definitions
This file automatically generated by lcm.
DO NOT MODIFY BY HAND!!!!
"""


from io import BytesIO
import struct

class lcmt_header(object):
    """
    This holds timestamp and a particular coordinate frame.
    This follows ROS's convention except the timestamp since
    utime has been widely used in Drake LCM.
    """

    __slots__ = ["seq", "utime", "frame_name"]

    __typenames__ = ["int32_t", "int64_t", "string"]

    __dimensions__ = [None, None, None]

    def __init__(self):
        self.seq = 0
        """
        Sequence ID: consecutively increasing ID.
        LCM Type: int32_t
        """

        self.utime = 0
        """
        Timestamp in microseconds.
        LCM Type: int64_t
        """

        self.frame_name = ""
        """
        The name of a frame this data is associated with.
        LCM Type: string
        """


    def encode(self):
        buf = BytesIO()
        buf.write(lcmt_header._get_packed_fingerprint())
        self._encode_one(buf)
        return buf.getvalue()

    def _encode_one(self, buf):
        buf.write(struct.pack(">iq", self.seq, self.utime))
        __frame_name_encoded = self.frame_name.encode('utf-8')
        buf.write(struct.pack('>I', len(__frame_name_encoded)+1))
        buf.write(__frame_name_encoded)
        buf.write(b"\0")

    @staticmethod
    def decode(data: bytes):
        if hasattr(data, 'read'):
            buf = data
        else:
            buf = BytesIO(data)
        if buf.read(8) != lcmt_header._get_packed_fingerprint():
            raise ValueError("Decode error")
        return lcmt_header._decode_one(buf)

    @staticmethod
    def _decode_one(buf):
        self = lcmt_header()
        self.seq, self.utime = struct.unpack(">iq", buf.read(12))
        __frame_name_len = struct.unpack('>I', buf.read(4))[0]
        self.frame_name = buf.read(__frame_name_len)[:-1].decode('utf-8', 'replace')
        return self

    @staticmethod
    def _get_hash_recursive(parents):
        if lcmt_header in parents: return 0
        tmphash = (0x9272c333198c72a) & 0xffffffffffffffff
        tmphash  = (((tmphash<<1)&0xffffffffffffffff) + (tmphash>>63)) & 0xffffffffffffffff
        return tmphash
    _packed_fingerprint = None

    @staticmethod
    def _get_packed_fingerprint():
        if lcmt_header._packed_fingerprint is None:
            lcmt_header._packed_fingerprint = struct.pack(">Q", lcmt_header._get_hash_recursive([]))
        return lcmt_header._packed_fingerprint

    def get_hash(self):
        """Get the LCM hash of the struct"""
        return struct.unpack(">Q", lcmt_header._get_packed_fingerprint())[0]




import numpy
import pydrake.geometry
import pydrake.multibody.math
import pydrake.multibody.plant
import pydrake.multibody.tree
import pydrake.solvers
import pydrake.symbolic
import pydrake.systems.framework
import pydrake.trajectories
from typing import ClassVar, overload

class CalcGridPointsOptions:
    __fields__: ClassVar[tuple] = ...  # read-only
    max_err: float
    max_iter: int
    max_seg_length: float
    min_points: int
    def __init__(self, **kwargs) -> None: ...
    def __copy__(self) -> CalcGridPointsOptions: ...
    def __deepcopy__(self, arg0: dict) -> CalcGridPointsOptions: ...

class CentroidalMomentumConstraint(pydrake.solvers.Constraint):
    def __init__(self, plant: pydrake.multibody.plant.MultibodyPlant_TAutoDiffXdU, model_instances: list[pydrake.multibody.tree.ModelInstanceIndex] | None, plant_context: pydrake.systems.framework.Context_TAutoDiffXdU, angular_only: bool) -> None: ...

class ContactWrench:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def F_Cb_W(self) -> pydrake.multibody.math.SpatialForce: ...
    @property
    def bodyA_index(self) -> pydrake.multibody.tree.BodyIndex: ...
    @property
    def bodyB_index(self) -> pydrake.multibody.tree.BodyIndex: ...
    @property
    def p_WCb_W(self) -> numpy.ndarray[numpy.float64[3, 1]]: ...

class ContactWrenchFromForceInWorldFrameEvaluator(pydrake.solvers.EvaluatorBase):
    def __init__(self, plant: pydrake.multibody.plant.MultibodyPlant_TAutoDiffXdU, context: pydrake.systems.framework.Context_TAutoDiffXdU, geometry_id_pair: tuple[pydrake.geometry.GeometryId]) -> None: ...

class QuaternionEulerIntegrationConstraint(pydrake.solvers.Constraint):
    def __init__(self, allow_quaternion_negation: bool) -> None: ...
    @overload
    def ComposeVariable(self, quat1: numpy.ndarray[numpy.float64[4, 1]], quat2: numpy.ndarray[numpy.float64[4, 1]], angular_vel: numpy.ndarray[numpy.float64[3, 1]], h: float) -> numpy.ndarray[numpy.float64[12, 1]]: ...
    @overload
    def ComposeVariable(self, quat1: numpy.ndarray[object[4, 1]], quat2: numpy.ndarray[object[4, 1]], angular_vel: numpy.ndarray[object[3, 1]], h: pydrake.symbolic.Variable) -> numpy.ndarray[object[12, 1]]: ...
    @overload
    def ComposeVariable(self, quat1: numpy.ndarray[object[4, 1]], quat2: numpy.ndarray[object[4, 1]], angular_vel: numpy.ndarray[object[3, 1]], h: pydrake.symbolic.Expression) -> numpy.ndarray[object[12, 1]]: ...
    def allow_quaternion_negation(self) -> bool: ...

class SpatialVelocityConstraint(pydrake.solvers.Constraint):
    class AngularVelocityBounds:
        magnitude_lower: float
        magnitude_upper: float
        reference_direction: numpy.ndarray[numpy.float64[3, 1]]
        theta_bound: float
        def __init__(self) -> None: ...
    def __init__(self, plant: pydrake.multibody.plant.MultibodyPlant_TAutoDiffXdU, frameA: pydrake.multibody.tree.Frame_TAutoDiffXdU, v_AC_lower: numpy.ndarray[numpy.float64[3, 1]], v_AC_upper: numpy.ndarray[numpy.float64[3, 1]], frameB: pydrake.multibody.tree.Frame_TAutoDiffXdU, p_BCo: numpy.ndarray[numpy.float64[3, 1]], plant_context: pydrake.systems.framework.Context_TAutoDiffXdU, w_AC_bounds=...) -> None: ...

class StaticEquilibriumProblem:
    def __init__(self, plant: pydrake.multibody.plant.MultibodyPlant_TAutoDiffXdU, context: pydrake.systems.framework.Context_TAutoDiffXdU, ignored_collision_pairs: set[tuple[pydrake.geometry.GeometryId, pydrake.geometry.GeometryId]]) -> None: ...
    def GetContactWrenchSolution(self, result: pydrake.solvers.MathematicalProgramResult) -> list[ContactWrench]: ...
    def UpdateComplementarityTolerance(self, tol: float) -> None: ...
    def get_mutable_prog(self) -> pydrake.solvers.MathematicalProgram: ...
    def prog(self) -> pydrake.solvers.MathematicalProgram: ...
    def q_vars(self) -> numpy.ndarray[object[m, 1]]: ...
    def u_vars(self) -> numpy.ndarray[object[m, 1]]: ...

class Toppra:
    def __init__(self, path: pydrake.trajectories.Trajectory, plant: pydrake.multibody.plant.MultibodyPlant, gridpoints: numpy.ndarray[numpy.float64[m, 1]]) -> None: ...
    def AddFrameAccelerationLimit(self, constraint_frame: pydrake.multibody.tree.Frame, lower_limit: numpy.ndarray[numpy.float64[6, 1]], upper_limit: numpy.ndarray[numpy.float64[6, 1]], discretization: ToppraDiscretization = ...) -> tuple[pydrake.solvers.BindingTLinearConstraintU, pydrake.solvers.BindingTLinearConstraintU]: ...
    @overload
    def AddFrameTranslationalSpeedLimit(self, constraint_frame: pydrake.multibody.tree.Frame, upper_limit: float) -> pydrake.solvers.BindingTBoundingBoxConstraintU: ...
    @overload
    def AddFrameTranslationalSpeedLimit(self, constraint_frame: pydrake.multibody.tree.Frame, upper_limit: pydrake.trajectories.Trajectory) -> pydrake.solvers.BindingTBoundingBoxConstraintU: ...
    def AddFrameVelocityLimit(self, constraint_frame: pydrake.multibody.tree.Frame, lower_limit: numpy.ndarray[numpy.float64[6, 1]], upper_limit: numpy.ndarray[numpy.float64[6, 1]]) -> pydrake.solvers.BindingTBoundingBoxConstraintU: ...
    def AddJointAccelerationLimit(self, lower_limit: numpy.ndarray[numpy.float64[m, 1]], upper_limit: numpy.ndarray[numpy.float64[m, 1]], discretization: ToppraDiscretization = ...) -> tuple[pydrake.solvers.BindingTLinearConstraintU, pydrake.solvers.BindingTLinearConstraintU]: ...
    def AddJointTorqueLimit(self, lower_limit: numpy.ndarray[numpy.float64[m, 1]], upper_limit: numpy.ndarray[numpy.float64[m, 1]], discretization: ToppraDiscretization = ...) -> tuple[pydrake.solvers.BindingTLinearConstraintU, pydrake.solvers.BindingTLinearConstraintU]: ...
    def AddJointVelocityLimit(self, lower_limit: numpy.ndarray[numpy.float64[m, 1]], upper_limit: numpy.ndarray[numpy.float64[m, 1]]) -> pydrake.solvers.BindingTBoundingBoxConstraintU: ...
    @staticmethod
    def CalcGridPoints(path: pydrake.trajectories.Trajectory, options: CalcGridPointsOptions) -> numpy.ndarray[numpy.float64[m, 1]]: ...
    def SolvePathParameterization(self, s_dot_start: float = ..., s_dot_end: float = ...) -> pydrake.trajectories.PiecewisePolynomial | None: ...
    def constraint_relaxation(self) -> float: ...
    def set_constraint_relaxation(self, relaxation: float) -> None: ...

class ToppraDiscretization:
    __members__: ClassVar[dict] = ...  # read-only
    __entries: ClassVar[dict] = ...
    kCollocation: ClassVar[ToppraDiscretization] = ...
    kInterpolation: ClassVar[ToppraDiscretization] = ...
    def __init__(self, value: int) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...



"""
Run examples of PlanarSceneGraphVisualizer, e.g. to visualize a pendulum.
Usage demo: load a URDF, rig it up with a constant torque input, and draw it.
"""

import argparse

import numpy as np

from pydrake.common import configure_logging, temp_directory
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import AddMultibodyPlantSceneGraph
from pydrake.systems.analysis import Simulator
from pydrake.systems.framework import DiagramBuilder
from pydrake.systems.planar_scenegraph_visualizer import (
    ConnectPlanarSceneGraphVisualizer,
)


def run_pendulum_example(args):
    builder = DiagramBuilder()
    plant, scene_graph = AddMultibodyPlantSceneGraph(builder, 0.0)
    parser = Parser(builder)
    parser.AddModelsFromUrl(
        url="package://drake/examples/pendulum/Pendulum.urdf"
    )
    plant.Finalize()

    T_VW = np.array(
        [
            [1.0, 0.0, 0.0, 0.0],  # BR
            [0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )
    visualizer = ConnectPlanarSceneGraphVisualizer(
        builder, scene_graph, T_VW=T_VW, xlim=[-1.2, 1.2], ylim=[-1.2, 1.2]
    )

    if args.playback:
        visualizer.start_recording()

    diagram = builder.Build()
    simulator = Simulator(diagram)
    simulator.Initialize()
    simulator.set_target_realtime_rate(1.0)

    # Fix the input port to zero.
    plant_context = diagram.GetMutableSubsystemContext(
        plant, simulator.get_mutable_context()
    )
    plant.get_actuation_input_port().FixValue(
        plant_context, np.zeros(plant.num_actuators())
    )
    plant_context.SetContinuousState([0.5, 0.1])
    simulator.AdvanceTo(args.duration)

    if args.playback:
        visualizer.stop_recording()
        ani = visualizer.get_recording_as_animation()
        # Playback the recording and save the output.
        ani.save(f"{temp_directory()}/pend_playback.mp4", fps=30)


def main():
    configure_logging()
    np.set_printoptions(precision=5, suppress=True)
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-T",
        "--duration",
        type=float,
        help="Duration to run sim in seconds.",
        default=1.0,
    )
    parser.add_argument(
        "-p",
        "--playback",
        type=bool,
        help="Whether to record and playback the animation",
        default=False,
    )
    args = parser.parse_args()
    run_pendulum_example(args)


if __name__ == "__main__":
    main()



# Remove once we have Python >= 3.10.
from __future__ import annotations

import base64
import copy
import hashlib
import json
import logging
from pathlib import Path
import re
import sys
import time

import numpy as np

from drake import (
    lcmt_contact_results_for_viz,
    lcmt_point_cloud,
    lcmt_point_cloud_field,
    lcmt_viewer_draw,
    lcmt_viewer_geometry_data,
    lcmt_viewer_link_data,
    lcmt_viewer_load_robot,
)
from pydrake.common import MemoryFile
from pydrake.common.eigen_geometry import (
    Quaternion,
)
from pydrake.geometry import (
    Box,
    Capsule,
    Cylinder,
    Ellipsoid,
    InMemoryMesh,
    Mesh,
    Meshcat,
    MeshcatParams,
    MeshSource,
    Rgba,
    Sphere,
    SurfaceTriangle,
    TriangleSurfaceMesh,
)
from pydrake.lcm import (
    DrakeLcm,
)
from pydrake.math import (
    RigidTransform,
    RotationMatrix,
)
from pydrake.multibody.meshcat import (
    ContactVisualizerParams,
    _HydroelasticContactVisualizer,
    _HydroelasticContactVisualizerItem,
    _PointContactVisualizer,
    _PointContactVisualizerItem,
)
from pydrake.perception import (
    BaseField,
    Fields,
    PointCloud,
)

_logger = logging.getLogger("drake")

_DEFAULT_MESHCAT_PARAMS = MeshcatParams(
    host="localhost",
    show_stats_plot=True,
)


def _to_pose(position, quaternion):
    """Given pose parts, parses it into a RigidTransform."""
    return RigidTransform(Quaternion(wxyz=quaternion), p=position)


class _Slider:
    """A slider with range [small-positive-value to 1.0]."""

    def __init__(self, meshcat, name):
        self._meshcat = meshcat
        self._name = name
        self._value = 1.0
        self._exists = False

    def realize(self):
        """Ensures the slider is created and visible."""
        # Avoid a redundant slider crash if the user runs a sending program
        # twice with the same meldis still running.
        if not self._exists:
            self._exists = True
            self._meshcat.AddSlider(self._name, 0.02, 1.0, 0.02, self._value)

    def read(self):
        """Returns a pair of the current value, and a flag that is True if the
        value has changed since the last read().

        Clients *must* call realize() before calling read().
        """
        assert self._exists, "Call realize() before calling read()"
        value = self._meshcat.GetSliderValue(self._name)
        value_changed = value != self._value
        self._value = value
        return value, value_changed


def _json_to_memory_file(json):
    """Converts a json representation of a MemoryFile to an instance of
    same."""
    return MemoryFile(
        contents=base64.b64decode(json["contents"]),
        extension=json["extension"],
        filename_hint=json["filename_hint"],
    )


def _json_to_file_source(json):
    """Converts the json representation of a file source to an instance of
    same."""
    if "path" in json:
        return json["path"]
    else:
        return _json_to_memory_file(json)


def _make_mesh_source_from_json(json_str):
    """Given the json representation of a mesh, create a mesh source for it."""
    # Generally, we assume that if it is legitimate json, then it is an
    # an in-memory representation.
    try:
        payload = json.loads(json_str)
    except json.JSONDecodeError as e:
        _logger.warning(f"Received message with malformed json: {e}")
        return None

    if (
        "in_memory_mesh" not in payload
        or "mesh_file" not in payload["in_memory_mesh"]
    ):
        _logger.warning("Received Mesh with unexpected json content")
        return None

    mesh_data = payload["in_memory_mesh"]
    supporting_files = {}
    for name, coded in mesh_data.get("supporting_files", {}).items():
        supporting_files[name] = _json_to_file_source(coded)

    return MeshSource(
        InMemoryMesh(
            mesh_file=_json_to_memory_file(mesh_data["mesh_file"]),
            supporting_files=supporting_files,
        )
    )


class _GeometryFileHasher:
    """Calculates a checksum of external file(s) referenced by geometry
    messages such as lcmt_viewer_load_robot or similar.

    Each "on_..." method incorporates all of the external files cited by the
    given argument (of a specific type) into the current hash. Files that are
    named but cannot be opened are silently skipped.
    """

    def __init__(self):
        self._paths = []
        self._hasher = hashlib.sha256()

    def value(self):
        return self._hasher.hexdigest()

    def _read_file(self, path: Path):
        """Reads the given file and adds its content to the current hash.
        Returns the file content (as ``bytes``, not ``str``).
        Remembers the filename (for unit testing).
        If the file is missing, silently returns an empty ``bytes``.
        """
        try:
            with open(path, "rb") as f:
                content = f.read()
            self._hasher.update(content)
            self._paths.append(path)
            return content
        except OSError:
            return b""

    def on_viewer_load_robot(self, message: lcmt_viewer_load_robot):
        assert isinstance(message, lcmt_viewer_load_robot)
        for link in message.link:
            for geom in link.geom:
                self.on_viewer_geometry_data(geom)

    def on_viewer_geometry_data(self, message: lcmt_viewer_geometry_data):
        assert isinstance(message, lcmt_viewer_geometry_data)
        if (
            message.type == lcmt_viewer_geometry_data.MESH
            and message.string_data
        ):
            self.on_mesh(message.string_data)

    def on_mesh(self, string_data: str):
        if string_data[0] == "{":
            self.on_mesh_in_memory(string_data)
        else:
            self.on_mesh_from_disk(Path(string_data))

    def on_mesh_in_memory(self, string_data: str):
        json_str = string_data.encode()
        source = _make_mesh_source_from_json(json_str)
        if source is None:
            # While the json won't produce a source that will instantiate
            # geometry in the _ViewerApplet, we'll hash the contents of the
            # json so that if it's corrected, we'll redraw on it.
            self._hasher.update(json_str)
            return

        assert source.is_in_memory()

        # Note: we only need to hash *supporting* files that are defined by
        # their path. The main mesh file and all in-memory supporting files
        # appear in the message body itself, and, as such, changes to that data
        # gets detected *outside* of the hasher.
        for file_source in source.in_memory().supporting_files.values():
            if isinstance(file_source, Path):
                self._read_file(file_source)

    def on_mesh_from_disk(self, path: Path):
        # Hash the file contents, even if we don't know how to interpret it.
        content = self._read_file(path)
        if path.suffix.lower() == ".obj":
            self.on_obj_from_disk(path, content)
        elif path.suffix.lower() == ".gltf":
            self.on_gltf_from_disk(path, content)
        else:
            _logger.warning(
                f"Unsupported mesh file: '{path}'\n"
                "Update Meldis's hasher to trigger reloads on this kind of "
                "file."
            )

    def on_obj_from_disk(self, path: Path, content: bytes):
        assert isinstance(path, Path)
        for mtl_names in re.findall(
            rb"^\s*mtllib\s+(.*?)\s*$", content, re.MULTILINE
        ):
            for mtl_name in mtl_names.decode("utf-8").split():
                self.on_mtl_from_disk(path.parent / mtl_name)

    def on_mtl_from_disk(self, path: Path):
        assert isinstance(path, Path)
        content = self._read_file(path)
        for tex_name in re.findall(
            rb"^\s*map_.*?\s+(\S+)\s*$", content, re.MULTILINE
        ):
            self.on_texture_from_disk(path.parent / tex_name.decode("utf-8"))

    def on_texture_from_disk(self, path: Path):
        assert isinstance(path, Path)
        self._read_file(path)

    def on_gltf_from_disk(self, path: Path, content: bytes):
        assert isinstance(path, Path)
        try:
            document = json.loads(content.decode(encoding="utf-8"))
        except json.JSONDecodeError:
            _logger.warning(f"glTF file is not valid JSON: {path}")
            return

        # Handle images and .bin files cited via URIs.
        for array_property in ("images", "buffers"):
            for item in document.get(array_property, []):
                uri = item.get("uri", None)
                if uri and not uri.startswith("data:"):
                    self._read_file(path.parent / uri)


class _ViewerApplet:
    """Displays lcmt_viewer_load_robot and lcmt_viewer_draw into MeshCat."""

    def __init__(
        self,
        *,
        meshcat,
        path,
        alpha_slider_name,
        should_accept_link=None,
        start_visible=True,
        initial_alpha_value=1,
    ):
        """Constructs an applet.

        If should_accept_link is given, only links where
        `should_accept_link(link.name)` is true will be displayed. (N.B. This
        predicate only applies to rigid bodies; it does not apply to deformable
        geometries, which do not have a concept of a "link".)
        """
        self._meshcat = meshcat
        self._path = path
        self._load_message = None
        self._load_message_mesh_checksum = None
        self._alpha_slider = _Slider(meshcat, f"{alpha_slider_name} α")
        self._alpha_slider._value = initial_alpha_value
        if should_accept_link is not None:
            self._should_accept_link = should_accept_link
        else:
            self._should_accept_link = lambda _: True
        self._applet_name = alpha_slider_name
        self._start_visible = start_visible
        self._geom_paths = []

        # Initialize ourself with an empty load message.
        self.on_viewer_load(message=lcmt_viewer_load_robot())

    def on_viewer_load(self, message):
        """Handler for lcmt_viewer_load."""
        # Ignore duplicate load messages. This is important for visualization
        # performance when the user is repeatedly viewing the same simulation
        # over and over again, since reloading a scene into Meshcat has high
        # latency.
        hasher = _GeometryFileHasher()
        hasher.on_viewer_load_robot(message)
        mesh_checksum = hasher.value()
        if self._load_message is not None:
            if (
                message.num_links == self._load_message.num_links
                and message.encode() == self._load_message.encode()
                and mesh_checksum == self._load_message_mesh_checksum
            ):
                _logger.info(
                    f"Ignoring duplicate load message for {self._applet_name}."
                )
                return

        # The semantics of a load message is to reset the entire scene.
        self._meshcat.Delete(path=self._path)

        self._waiting_for_first_draw_message = True
        self._load_message = message
        self._load_message_mesh_checksum = mesh_checksum
        self.set_alpha(self._alpha_slider._value, on_load=True)

    def _build_links(self):
        # Make all of our (ViewerApplet's) geometry invisible so that the
        # lcmt_viewer_load geometry does not clutter up the scene until we
        # are given its poses in a lcmt_viewer_draw message.
        self._set_visible(False)

        message = self._load_message

        # Add the links and their geometries.
        self._geom_paths = []
        for link in message.link:
            if not self._should_accept_link(link.name):
                continue
            robot_num = link.robot_num
            link_name = link.name.replace("::", "/")
            link_path = f"{self._path}/{robot_num}/{link_name}"
            for j, geom in enumerate(link.geom):
                geom_path = f"{link_path}/{j}"
                shape, rgba, pose = self._convert_geom(geom)
                if shape is None:
                    continue
                self._geom_paths.append(geom_path)
                set_object_kwargs = dict(path=geom_path, rgba=rgba)
                if isinstance(shape, TriangleSurfaceMesh):
                    set_object_kwargs.update(mesh=shape)
                else:
                    set_object_kwargs.update(shape=shape)
                self._meshcat.SetObject(**set_object_kwargs)
                self._meshcat.SetTransform(path=geom_path, X_ParentPath=pose)

    def on_viewer_draw(self, message):
        """Handler for lcmt_viewer_draw."""
        for i in range(message.num_links):
            if not self._should_accept_link(message.link_name[i]):
                continue
            link_name = message.link_name[i].replace("::", "/")
            robot_num = message.robot_num[i]
            link_path = f"{self._path}/{robot_num}/{link_name}"
            pose = _to_pose(message.position[i], message.quaternion[i])
            self._meshcat.SetTransform(path=link_path, X_ParentPath=pose)
        self._meshcat.SetSimulationTime(sim_time=message.timestamp * 1e-3)
        if self._waiting_for_first_draw_message:
            self._waiting_for_first_draw_message = False
            self._build_links()
            self._alpha_slider.realize()
            if self._start_visible:
                self._set_visible(True)
            self.on_poll(force=True)

    def set_alpha(self, value, on_load: bool):
        """Applies the given alpha `value` to the visualized geometries."""
        if on_load:
            # When we're loading the geometries, we have to address each
            # geometry individually because they may not be loaded yet in the
            # browser.
            for path in self._geom_paths:
                self._meshcat.SetProperty(path, "modulated_opacity", value)
        else:
            # When things are truly loaded and we're reacting to changes in
            # the slider value, we can simply set the root and let meshcat.js
            # percolate the change through the tree.
            self._meshcat.SetProperty(self._path, "modulated_opacity", value)

    def on_poll(self, force=False):
        if self._waiting_for_first_draw_message:
            return
        value, value_changed = self._alpha_slider.read()
        if force or value_changed:
            self.set_alpha(value, on_load=False)

    def on_viewer_draw_deformable(self, message):
        """Handler for lcmt_viewer_link_data."""
        link_name = message.name
        robot = message.robot_num
        link_path = f"{self._path}/{robot}/{link_name}"
        for i, geom in enumerate(message.geom):
            geom_name = geom.string_data
            geom_path = f"{link_path}/{geom_name}"
            vertices, faces, rgba, pose = self._convert_deformable_geom(geom)
            self._meshcat.SetTriangleMesh(
                path=geom_path, vertices=vertices, faces=faces, rgba=rgba
            )
            self._meshcat.SetTransform(path=link_path, X_ParentPath=pose)
        if self._waiting_for_first_draw_message:
            self._waiting_for_first_draw_message = False
            self._set_visible(True)

    def _set_visible(self, value):
        self._meshcat.SetProperty(self._path, property="visible", value=value)

    def _convert_deformable_geom(self, geom):
        """Given an lcmt_viewer_geometry_data, parses it into a tuple of
        (vertices, faces, Rgba, RigidTransform) if the geometry type is a
        MESH.
        """
        assert geom.type == lcmt_viewer_geometry_data.MESH
        num_verts = int(geom.float_data[0])
        num_faces = int(geom.float_data[1])
        # The first two floats encode the number of vertices and number of
        # triangles.
        v_start_index = 2
        f_start_index = v_start_index + 3 * num_verts
        vertices = np.array(geom.float_data[v_start_index:f_start_index])
        faces = np.array(geom.float_data[f_start_index:]).astype(int)
        vertices = np.reshape(vertices, (3, num_verts), order="F")
        faces = np.reshape(faces, (3, num_faces), order="F")
        rgba = Rgba(*geom.color)
        pose = _to_pose(geom.position, geom.quaternion)
        return (vertices, faces, rgba, pose)

    def _convert_geom(self, geom):
        """Given an lcmt_viewer_geometry_data, parses it into a tuple of
        (Shape, Rgba, RigidTransform) or
        (TriangleSurfaceMesh, Rgbd, RigidTransform).
        """
        shape = None
        if geom.type == lcmt_viewer_geometry_data.BOX:
            (width, depth, height) = geom.float_data
            shape = Box(width=width, depth=depth, height=height)
        elif geom.type == lcmt_viewer_geometry_data.CAPSULE:
            (radius, length) = geom.float_data
            shape = Capsule(radius=radius, length=length)
        elif geom.type == lcmt_viewer_geometry_data.CYLINDER:
            (radius, length) = geom.float_data
            shape = Cylinder(radius=radius, length=length)
        elif geom.type == lcmt_viewer_geometry_data.ELLIPSOID:
            (a, b, c) = geom.float_data
            shape = Ellipsoid(a=a, b=b, c=c)
        elif geom.type == lcmt_viewer_geometry_data.MESH and geom.string_data:
            scale3 = geom.float_data
            if geom.string_data[0] == "{":
                mesh_source = _make_mesh_source_from_json(geom.string_data)
                if mesh_source is None:
                    # Warning has already been emitted by _make_mesh_source...
                    return (None, None, None)
                shape = Mesh(source=mesh_source, scale3=scale3)
            else:
                # A mesh to be loaded from a file.
                filename = geom.string_data
                shape = Mesh(filename=filename, scale3=scale3)
        elif geom.type == lcmt_viewer_geometry_data.MESH:
            assert not geom.string_data
            shape = self._make_triangle_mesh(geom.float_data)
        elif geom.type == lcmt_viewer_geometry_data.SPHERE:
            (radius,) = geom.float_data
            shape = Sphere(radius=radius)
        else:
            _logger.warning(f"Unknown geom.type of {geom.type}")
            return (None, None, None)
        rgba = Rgba(*geom.color)
        pose = _to_pose(geom.position, geom.quaternion)
        return (shape, rgba, pose)

    @staticmethod
    def _make_triangle_mesh(data):
        """Returns a TriangleSurfaceMesh parsed from the given float data.
        The data is formatted like this:
          V | T | v0 | v1 | ... vN | t0 | t1 | ... | tM
        where
          V: The number of vertices.
          T: The number of triangles.
          N: 3V, the number of floating point values for the V vertices.
          M: 3T, the number of vertex indices for the T triangles.
        """
        V = int(data[0] if len(data) > 0 else -1)
        T = int(data[1] if len(data) > 1 else -1)
        if len(data) != 2 + 3 * V + 3 * T:
            _logger.warning("Ignoring mesh with malformed data length.")
            return None
        start = 2
        vertices = []
        for i in range(start, start + 3 * V, 3):
            vertex = np.array(data[i : i + 3])
            vertices.append(vertex)
        start = 2 + 3 * V
        triangles = []
        for i in range(start, start + 3 * T, 3):
            indices = [int(j) for j in data[i : i + 3]]
            triangle = SurfaceTriangle(*indices)
            triangles.append(triangle)
        return TriangleSurfaceMesh(triangles=triangles, vertices=vertices)


class _ContactApplet:
    """Displays lcmt_contact_results_for_viz into Meshcat."""

    def __init__(self, *, meshcat):
        # By default, don't show any contact illustrations.
        meshcat.SetProperty("/CONTACT_RESULTS", "visible", True)

        # Add point visualization.
        params = ContactVisualizerParams()
        params.prefix = "/CONTACT_RESULTS/point"
        self._point_helper = _PointContactVisualizer(meshcat, params)

        # Add hydroelastic visualization.
        params = ContactVisualizerParams()
        params.prefix = "/CONTACT_RESULTS/hydroelastic"
        self._hydro_helper = _HydroelasticContactVisualizer(meshcat, params)

    # Converts poly_data from a hydro lcm message to numpy array.
    def convert_faces(self, poly_data):
        poly_index = 0
        faces = []
        while poly_index < len(poly_data):
            poly_i_num_vertices = poly_data[poly_index]
            vertex_0_index = poly_index + 1
            vertex_0_global_index = poly_data[vertex_0_index]
            for i in range(1, poly_i_num_vertices - 1):
                vertex_1_global_index = poly_data[vertex_0_index + i]
                vertex_2_global_index = poly_data[vertex_0_index + i + 1]
                faces.append(
                    [
                        vertex_0_global_index,
                        vertex_1_global_index,
                        vertex_2_global_index,
                    ]
                )
            poly_index += poly_i_num_vertices + 1
        return np.array(faces).transpose()

    # Converts verts from hydro lcm message to numpy array
    def convert_verts(self, p_WV):
        verts = np.empty((3, len(p_WV)))
        for i in range(len(p_WV)):
            verts[0, i] = p_WV[i].x
            verts[1, i] = p_WV[i].y
            verts[2, i] = p_WV[i].z
        return verts

    def get_full_names(self, item):
        name1 = []
        name2 = []

        # Use model instance name if necessary.
        if not item.body1_unique:
            name1.append(item.model1_name)

        if not item.body2_unique:
            name2.append(item.model2_name)

        name1.append(item.body1_name)
        name2.append(item.body2_name)

        # Use geometry name if necessary.
        if item.collision_count1 > 1:
            name1.append(item.geometry1_name)

        if item.collision_count2 > 1:
            name2.append(item.geometry2_name)

        return (".".join(name1), ".".join(name2))

    def on_contact_results(self, message):
        """Handler for lcmt_contact_results_for_viz. Note that only point
        hydroelastic contact force and moment vectors are shown; contact
        surface and pressure are not shown.
        """

        # Handle point contact pairs
        viz_items = []
        for lcm_item in message.point_pair_contact_info:
            viz_items.append(
                _PointContactVisualizerItem(
                    body_A=lcm_item.body1_name,
                    body_B=lcm_item.body2_name,
                    contact_force=lcm_item.contact_force,
                    contact_point=lcm_item.contact_point,
                )
            )
        self._point_helper.Update(0, viz_items)

        # Handle hydroelastic contact pairs
        viz_items = []
        for lcm_item in message.hydroelastic_contacts:
            (name1, name2) = self.get_full_names(lcm_item)
            viz_items.append(
                _HydroelasticContactVisualizerItem(
                    body_A=name1,
                    body_B=name2,
                    centroid_W=lcm_item.centroid_W,
                    force_C_W=lcm_item.force_C_W,
                    moment_C_W=lcm_item.moment_C_W,
                    p_WV=self.convert_verts(lcm_item.p_WV),
                    faces=self.convert_faces(lcm_item.poly_data),
                    pressure=lcm_item.pressure,
                )
            )
        self._hydro_helper.Update(0, viz_items)


class _PointCloudApplet:
    """Displays lcmt_point_cloud into MeshCat."""

    _POINT_CLOUD_FIELDS = (
        # (name, byte_offset, datatype, count)
        ("x", 0, lcmt_point_cloud_field.FLOAT32, 1),
        ("y", 4, lcmt_point_cloud_field.FLOAT32, 1),
        ("z", 8, lcmt_point_cloud_field.FLOAT32, 1),
        ("rgb", 12, lcmt_point_cloud_field.UINT32, 1),
        ("normal_x", 16, lcmt_point_cloud_field.FLOAT32, 1),
        ("normal_y", 20, lcmt_point_cloud_field.FLOAT32, 1),
        ("normal_z", 24, lcmt_point_cloud_field.FLOAT32, 1),
    )
    """The supported fields and their data types of a point cloud. An XYZ,
    XYZRGB, and XYZRGBNormal cloud will have the first three, first four, and
    all the fields in this particular order.
    """

    def __init__(self, *, meshcat):
        self._meshcat = meshcat
        self._already_warned_channel_names = set()

    def _validate_and_get_fields(self, message):
        """Checks the point cloud LCM message and returns the corresponding
        `Fields` for the PointCloud object. Either XYZ, XYZRGB, or XYZRGBNormal
        cloud with the exact data type is supported.
        """
        if message.flags != lcmt_point_cloud.IS_STRICTLY_FINITE:
            return None

        if message.num_fields not in (3, 4, 7):
            return None

        for i in range(message.num_fields):
            (name, byte_offset, datatype, count) = self._POINT_CLOUD_FIELDS[i]
            if (
                message.fields[i].name != name
                or message.fields[i].byte_offset != byte_offset
                or message.fields[i].datatype != datatype
                or message.fields[i].count != count
            ):
                return None

        if message.num_fields == 3:
            return Fields(BaseField.kXYZs)
        elif message.num_fields == 4:
            return Fields(BaseField.kXYZs | BaseField.kRGBs)
        else:
            return Fields(
                BaseField.kXYZs | BaseField.kRGBs | BaseField.kNormals
            )

    def _channel_to_meshcat_path(self, channel):
        assert channel.startswith("DRAKE_POINT_CLOUD")
        if channel == "DRAKE_POINT_CLOUD":
            return "/POINT_CLOUD/default"
        else:
            # E.g., `DRAKE_POINT_CLOUD_FOO` => `/POINT_CLOUD/FOO`.
            suffix = channel[len("DRAKE_POINT_CLOUD_") :]
            return f"/POINT_CLOUD/{suffix}"

    def on_point_cloud(self, channel, message):
        """Handler for lcmt_point_cloud.

        Validates and converts the lcmt_point_cloud message to a PointCloud
        object for display.
        """
        cloud_fields = self._validate_and_get_fields(message)
        if cloud_fields is None:
            # Throttle warning messages to one per channel.
            if channel not in self._already_warned_channel_names:
                self._already_warned_channel_names.add(channel)
                _logger.warning(f"Unsupported point cloud data from {channel}.")
            return

        # Transform the raw data into an N x num_fields array.
        raw_data = np.frombuffer(message.data, dtype=np.float32).reshape(
            -1, message.num_fields
        )
        num_points = raw_data.shape[0]

        cloud = PointCloud(num_points, cloud_fields)
        xyzs = raw_data[:, 0:3]
        cloud.mutable_xyzs()[:] = xyzs.transpose()
        if message.num_fields > 3:
            rgbs_with_padding = (
                raw_data[:, 3].astype(np.float32).view(np.uint8).reshape(-1, 4)
            )
            rgbs = rgbs_with_padding[:, 0:3]
            cloud.mutable_rgbs()[:] = rgbs.transpose()
        if message.num_fields > 4:
            normals = raw_data[:, 4:]
            cloud.mutable_normals()[:] = normals.transpose()

        self._meshcat.SetObject(
            path=self._channel_to_meshcat_path(channel),
            cloud=cloud,
            point_size=0.01,
        )


class _DrawFrameApplet:
    """Applet to visualize triads in meshcat"""

    def __init__(self, *, meshcat):
        """Constructs an applet."""
        self._meshcat = meshcat
        # previously published link names
        self._channel_link_map = {}

    def _channel_to_meshcat_path(self, channel):
        assert channel.startswith("DRAKE_DRAW_FRAMES")
        if channel == "DRAKE_DRAW_FRAMES":
            return "/DRAKE_DRAW_FRAMES/default"
        else:
            # E.g., `DRAKE_DRAW_FRAMES_FOO` => `/DRAKE_DRAW_FRAMES/FOO`.
            suffix = channel[len("DRAKE_DRAW_FRAMES_") :]
            return f"/DRAKE_DRAW_FRAMES/{suffix}"

    def _add_meshcat_triad(self, path, X_PT):
        length = 0.25
        radius = 0.01
        opacity = 1.0
        self._meshcat.SetTransform(path, X_PT)
        # x-axis
        X_TG = RigidTransform(
            RotationMatrix.MakeYRotation(np.pi / 2), [length / 2.0, 0, 0]
        )
        self._meshcat.SetTransform(path + "/x-axis", X_TG)
        self._meshcat.SetObject(
            path + "/x-axis", Cylinder(radius, length), Rgba(1, 0, 0, opacity)
        )
        # y-axis
        X_TG = RigidTransform(
            RotationMatrix.MakeXRotation(np.pi / 2), [0, length / 2.0, 0]
        )
        self._meshcat.SetTransform(path + "/y-axis", X_TG)
        self._meshcat.SetObject(
            path + "/y-axis", Cylinder(radius, length), Rgba(0, 1, 0, opacity)
        )
        # z-axis
        X_TG = RigidTransform([0, 0, length / 2.0])
        self._meshcat.SetTransform(path + "/z-axis", X_TG)
        self._meshcat.SetObject(
            path + "/z-axis", Cylinder(radius, length), Rgba(0, 0, 1, opacity)
        )

    def on_frame_update(self, channel, message):
        """Handler to update triads in meshcat. It updates poses sent using
        the lcmt_viewer_draw message."""
        channel_path = self._channel_to_meshcat_path(channel)

        # delete old frames if the link names have changed
        link_names = set(message.link_name)
        if self._channel_link_map.get(channel) != link_names:
            self._meshcat.Delete(path=channel_path)
        self._channel_link_map[channel] = link_names

        for i in range(message.num_links):
            link_name = message.link_name[i].replace("::", "/")
            link_path = f"{channel_path}/{link_name}"
            self._add_meshcat_triad(
                path=link_path,
                X_PT=_to_pose(message.position[i], message.quaternion[i]),
            )


class Meldis:
    """
    MeshCat LCM Display Server (MeLDiS)

    Offers a MeshCat visualization server that listens for and draws Drake's
    legacy LCM visualization messages.

    If the meshcat_host parameter is not supplied, 'localhost' will be used by
    default.

    Refer to the pydrake.visualization.meldis module docs for details.
    """

    def __init__(
        self,
        *,
        meshcat_host: str | None = None,
        meshcat_port: int | None = None,
        meshcat_params: MeshcatParams | None = None,
        environment_map: Path | None = None,
    ):
        """Constructs a new Meldis instance. The meshcat_host (when given)
        takes precedence over meshcat_params.host. The meshcat_post (when
        given) takes precedence over meshcat_params.port.
        """

        # Bookkeeping for update throttling.
        self._last_update_time = time.time()

        # Bookkeeping for subscriptions, keyed by LCM channel name.
        self._message_types = {}
        self._message_handlers = {}
        self._message_pending_data = {}

        self._poll_handlers = []

        self._lcm = DrakeLcm()
        lcm_url = self._lcm.get_lcm_url()
        _logger.info(f"Meldis is listening for LCM messages at {lcm_url}")

        # Create our meshcat object, merging all of the params goop into one.
        if meshcat_params is None:
            meshcat_params = _DEFAULT_MESHCAT_PARAMS
        params = copy.deepcopy(meshcat_params)
        if meshcat_host is not None:
            params.host = meshcat_host
        if meshcat_port is not None:
            params.port = meshcat_port
        self.meshcat = Meshcat(params=params)
        if environment_map is not None:
            self.meshcat.SetEnvironmentMap(environment_map)

        def is_inertia_link(link_name):
            return "::InertiaVisualizer::" in link_name

        def is_not_inertia_link(link_name):
            return not is_inertia_link(link_name)

        default_viewer = _ViewerApplet(
            meshcat=self.meshcat,
            path="/DRAKE_VIEWER",
            alpha_slider_name="Viewer",
            should_accept_link=is_not_inertia_link,
        )
        self._subscribe(
            channel="DRAKE_VIEWER_LOAD_ROBOT",
            message_type=lcmt_viewer_load_robot,
            handler=default_viewer.on_viewer_load,
        )
        self._subscribe(
            channel="DRAKE_VIEWER_DRAW",
            message_type=lcmt_viewer_draw,
            handler=default_viewer.on_viewer_draw,
        )
        self._subscribe(
            channel="DRAKE_VIEWER_DEFORMABLE",
            message_type=lcmt_viewer_link_data,
            handler=default_viewer.on_viewer_draw_deformable,
        )
        self._poll(handler=default_viewer.on_poll)

        inertia_viewer = _ViewerApplet(
            meshcat=self.meshcat,
            path="/Inertia Visualizer",
            alpha_slider_name="Inertia",
            should_accept_link=is_inertia_link,
            start_visible=False,
        )
        inertia_viewer._alpha_slider._value = 0.5
        self._subscribe(
            channel="DRAKE_VIEWER_LOAD_ROBOT",
            message_type=lcmt_viewer_load_robot,
            handler=inertia_viewer.on_viewer_load,
        )
        self._subscribe(
            channel="DRAKE_VIEWER_DRAW",
            message_type=lcmt_viewer_draw,
            handler=inertia_viewer.on_viewer_draw,
        )
        self._poll(handler=inertia_viewer.on_poll)

        illustration_viewer = _ViewerApplet(
            meshcat=self.meshcat,
            path="/Visual Geometry",
            alpha_slider_name="Visual",
        )
        self._subscribe(
            channel="DRAKE_VIEWER_LOAD_ROBOT_ILLUSTRATION",
            message_type=lcmt_viewer_load_robot,
            handler=illustration_viewer.on_viewer_load,
        )
        self._subscribe(
            channel="DRAKE_VIEWER_DRAW_ILLUSTRATION",
            message_type=lcmt_viewer_draw,
            handler=illustration_viewer.on_viewer_draw,
        )
        self._subscribe(
            channel="DRAKE_VIEWER_DEFORMABLE_ILLUSTRATION",
            message_type=lcmt_viewer_link_data,
            handler=illustration_viewer.on_viewer_draw_deformable,
        )
        self._poll(handler=illustration_viewer.on_poll)

        proximity_viewer = _ViewerApplet(
            meshcat=self.meshcat,
            path="/Collision Geometry",
            alpha_slider_name="Collision",
            start_visible=False,
            initial_alpha_value=0.5,
        )
        self._subscribe(
            channel="DRAKE_VIEWER_LOAD_ROBOT_PROXIMITY",
            message_type=lcmt_viewer_load_robot,
            handler=proximity_viewer.on_viewer_load,
        )
        self._subscribe(
            channel="DRAKE_VIEWER_DRAW_PROXIMITY",
            message_type=lcmt_viewer_draw,
            handler=proximity_viewer.on_viewer_draw,
        )
        self._subscribe(
            channel="DRAKE_VIEWER_DEFORMABLE_PROXIMITY",
            message_type=lcmt_viewer_link_data,
            handler=proximity_viewer.on_viewer_draw_deformable,
        )
        self._poll(handler=proximity_viewer.on_poll)

        contact = _ContactApplet(meshcat=self.meshcat)
        self._subscribe(
            channel="CONTACT_RESULTS",
            message_type=lcmt_contact_results_for_viz,
            handler=contact.on_contact_results,
        )

        # Subscribe to all the point-cloud-related channels.
        point_cloud = _PointCloudApplet(meshcat=self.meshcat)
        self._subscribe_multichannel(
            regex="DRAKE_POINT_CLOUD.*",
            message_type=lcmt_point_cloud,
            handler=point_cloud.on_point_cloud,
        )

        # Subscribe to all the frame display channels.
        draw_frame = _DrawFrameApplet(meshcat=self.meshcat)
        self._subscribe_multichannel(
            regex="DRAKE_DRAW_FRAMES.*",
            message_type=lcmt_viewer_draw,
            handler=draw_frame.on_frame_update,
        )

        # Bookkeeping for automatic shutdown.
        self._last_poll = None
        self._last_active = None

    def _subscribe(self, channel, message_type, handler):
        """Subscribes the handler to the given channel, using message_type to
        pass in a decoded message object (not the raw bytes). The handler will
        only be called at some maximum frequency. Messages on the same channel
        that arrive too quickly will be discarded.
        """
        # Record this channel's type and handler.
        assert self._message_types.get(channel, message_type) == message_type
        self._message_types[channel] = message_type

        # A wrapper to discard `channel` information as it's not used in the
        # actual handler.
        def _multi_handler(*, channel, message):
            handler(message)

        self._message_handlers.setdefault(channel, []).append(_multi_handler)

        # Subscribe using an internal function that implements "last one wins".
        # It's important to service the LCM queue as frequently as possible:
        #  https://github.com/RobotLocomotion/drake/issues/15234
        #  https://github.com/lcm-proj/lcm/issues/345
        # However, if the sender is transmitting visualization messages at
        # a high rate (e.g., if a sim is running much faster than realtime),
        # then we should only pass some of them along to MeshCat to avoid
        # flooding it. The handler merely records the message data; we'll
        # pass it along to MeshCat using our `self._should_update()` timer.
        def _on_message(data):
            self._message_pending_data[channel] = data

        self._lcm.Subscribe(channel=channel, handler=_on_message)

    def _subscribe_multichannel(self, regex, message_type, handler):
        """Subscribes the handler to a group of channels filtered by regex. How
        this function handles messages is the same as _subscribe() except that
        the channel name is only known when invoking the callback.
        """

        def _on_message(channel, data):
            if channel not in self._message_types:
                self._message_types[channel] = message_type
                self._message_handlers.setdefault(channel, []).append(handler)
            self._message_pending_data[channel] = data

        self._lcm.SubscribeMultichannel(regex=regex, handler=_on_message)

    def _poll(self, handler):
        self._poll_handlers.append(handler)

    def _invoke_poll(self):
        for function in self._poll_handlers:
            function()

    def _invoke_subscriptions(self):
        """Posts any unhandled messages to their handlers and clears the
        collection of unhandled messages.
        """
        for channel, data in self._message_pending_data.items():
            message = self._message_types[channel].decode(data)
            for function in self._message_handlers[channel]:
                function(channel=channel, message=message)

        self._message_pending_data.clear()

    def serve_forever(self, *, idle_timeout=None):
        """Runs indefinitely, forwarding LCM => MeshCat messages.

        If provided, the optional idle_timeout must be strictly positive and
        this loop will sys.exit after that many seconds without any websocket
        connections.
        """
        while True:
            self._lcm.HandleSubscriptions(timeout_millis=1000)
            if not self._should_update():
                continue
            self._invoke_subscriptions()
            self._invoke_poll()
            self.meshcat.Flush()
            self._check_for_shutdown(idle_timeout=idle_timeout)

    def _should_update(self):
        """Posts LCM-driven updates to MeshCat no faster than 40 Hz."""
        now = time.time()
        update_period = 0.025  # 40 Hz
        remaining = update_period - (now - self._last_update_time)
        if remaining > 0.0:
            return False
        else:
            self._last_update_time = now
            return True

    def _check_for_shutdown(self, *, idle_timeout):
        # Allow the user to opt-out of the timeout feature.
        if idle_timeout is None:
            return
        assert idle_timeout > 0.0

        # One-time initialization.
        now = time.time()
        if self._last_active is None:
            self._last_active = now
            return

        # Only check once every 5 seconds.
        if (self._last_poll is not None) and (now < self._last_poll + 5.0):
            return
        self._last_poll = now

        # Check to see if any browser client(s) are connected.
        if self.meshcat.GetNumActiveConnections() > 0:
            self._last_active = now
            return

        # In case we are idle for too long, exit automatically.
        if now > self._last_active + idle_timeout:
            _logger.info(
                "Meldis is exiting now; no browser was connected for"
                f" >{idle_timeout} seconds"
            )
            sys.exit(1)



"""
Defines a mapping between Python and alias types, and provides canonical
Python types as they relate to C++.
"""

import ctypes
import types
import typing

import numpy as np

from pydrake.common import _MangledName, pretty_class_name


class _StrictMap:
    # Provides a map which may only add a key once.
    def __init__(self):
        self._values = dict()

    def add(self, key, value):
        assert key not in self._values, f"Already added: {key}"
        self._values[key] = value

    def get(self, key, default):
        return self._values.get(key, default)


class _ParamAliases:
    # Registers aliases for a set of objects. This will be used for template
    # parameters.
    def __init__(self):
        self._to_canonical = _StrictMap()
        self._register_common()

    def _register_common(self):
        # Register common Python aliases relevant for C++.
        self.register(float, [np.double, ctypes.c_double])
        self.register(np.float32, [ctypes.c_float])
        self.register(int, [np.int32, ctypes.c_int32])
        self.register(np.int16, [ctypes.c_int16])
        self.register(np.int64, [ctypes.c_int64])
        self.register(np.uint8, [ctypes.c_uint8])
        self.register(np.uint16, [ctypes.c_uint16])
        self.register(np.uint32, [ctypes.c_uint32])
        self.register(np.uint64, [ctypes.c_uint64])
        # For Python versions < 3.14, convert typing.Union to types.UnionType.
        # This bring those older versions in line with how Python 3.14 works,
        # where *both* are a type (and aliases of each other).
        self.register(types.UnionType, [typing.Union])

    def register(self, canonical, aliases):
        # Registers a set of aliases to a canonical value.
        for alias in aliases:
            self._to_canonical.add(alias, canonical)

    def is_aliased(self, alias):
        # Determines if a parameter is aliased / registered.
        return self._to_canonical.get(alias, None) is not None

    def get_canonical(self, alias):
        # Gets registered canonical parameter if it is aliased; otherwise
        # return the same parameter.
        return self._to_canonical.get(alias, alias)

    @staticmethod
    def _resugar_typing_shortcuts(origin_name, arg_names):
        # Re-sugars typing.Union[T, NoneType] into typing.Optional[T] in case
        # that's what the origin and arg names refer to. Otherwise, returns the
        # data unchanged.
        if (
            origin_name == "typing.Union"
            and len(arg_names) == 2
            and arg_names[-1] == "NoneType"
        ):
            origin_name = "typing.Optional"
            arg_names = arg_names[:1]
        return origin_name, arg_names

    def get_name(self, alias, *, mangle):
        # Gets string for an alias.
        canonical = self.get_canonical(alias)
        if typing.get_origin(alias) is not None:
            origin = typing.get_origin(alias)
            args = typing.get_args(alias)
            origin_name = self.get_name(origin, mangle=mangle)
            arg_names = [self.get_name(x, mangle=mangle) for x in args]
            origin_name, arg_names = self._resugar_typing_shortcuts(
                origin_name, arg_names
            )
            result = f"{origin_name}[{','.join(arg_names)}]"
            if mangle:
                result = _MangledName.mangle(result)
            return result
        elif canonical is types.UnionType:
            return "typing.Union"
        elif isinstance(canonical, type):
            if mangle:
                return canonical.__name__
            else:
                return pretty_class_name(canonical)
        else:
            # For literals.
            result = str(canonical)
            if mangle:
                result = _MangledName.mangle(result)
            return result


# Create singleton instance.
_param_aliases = _ParamAliases()


def get_param_canonical(param):
    """Gets the canonical types for a set of Python types (canonical as in
    how they relate to C++ types).
    """
    return tuple(map(_param_aliases.get_canonical, param))


def get_param_names(param, *, mangle=False):
    """Gets the canonical type names for a set of Python types (canonical as
    in how they relate to C++ types).

    The ``mangle`` controls whether we use the nice name or the mangled name;
    see cpp_template.TemplateBase._instantiation_name for details.
    """
    return tuple(_param_aliases.get_name(x, mangle=mangle) for x in param)


class _Generic:
    """Provides a C++-compatible way to denote generic types. This uses the
    same type classes as Python's built-in generics (e.g., `typing.Union`) but
    is careful to canonicalize any type aliases in params during instantiation.
    """

    def __init__(self, name, instantiator, num_params):
        self._name = name
        self._instantiator = instantiator
        self._num_params = num_params

    def __getitem__(self, params):
        if not isinstance(params, tuple):
            params = (params,)
        if self._num_params is not None and len(params) != self._num_params:
            raise RuntimeError(
                f"{self._name}[] requires exactly "
                f"{self._num_params} type parameter(s)"
            )
        params_canonical = get_param_canonical(params)
        return self._instantiator(params_canonical)

    def __repr__(self):
        return f"<Generic {self._name}>"


def _dict_instantiator(params: tuple):
    (param_key, param_value) = params
    return dict[param_key, param_value]


def _list_instantiator(params: tuple):
    (param,) = params
    return list[param]


def _optional_instantiator(params: tuple):
    (param,) = params
    return typing.Optional[param]


def _union_instantiator(params: tuple):
    return typing.Union[params]


# A generic type `dict[KT, VT]` for the C++ class std::map<KT, VT>.
Dict = _Generic("Dict", _dict_instantiator, 2)

# A generic type `list[T]` for the C++ class std::vector<T>.
List = _Generic("List", _list_instantiator, 1)

# A generic type `typing.Optional[T]` for the C++ class std::optional<T>.
# Note that `typing` de-sugars this to be `typing.Union[T, NoneType]`.
Optional = _Generic("Optional", _optional_instantiator, 1)

# A generic type `typing.Union[X, ...]` for the C++ class std::variant<X, ...>.
Union = _Generic("Union", _union_instantiator, None)



"""LCM type definitions
This file automatically generated by lcm.
DO NOT MODIFY BY HAND!!!!
"""


from io import BytesIO
import struct

class lcmt_acrobot_y(object):

    __slots__ = ["timestamp", "theta1", "theta2", "tau"]

    __typenames__ = ["int64_t", "double", "double", "double"]

    __dimensions__ = [None, None, None, None]

    def __init__(self):
        self.timestamp = 0
        """ LCM Type: int64_t """
        self.theta1 = 0.0
        """ LCM Type: double """
        self.theta2 = 0.0
        """
        measured position
        LCM Type: double
        """

        self.tau = 0.0
        """ LCM Type: double """

    def encode(self):
        buf = BytesIO()
        buf.write(lcmt_acrobot_y._get_packed_fingerprint())
        self._encode_one(buf)
        return buf.getvalue()

    def _encode_one(self, buf):
        buf.write(struct.pack(">qddd", self.timestamp, self.theta1, self.theta2, self.tau))

    @staticmethod
    def decode(data: bytes):
        if hasattr(data, 'read'):
            buf = data
        else:
            buf = BytesIO(data)
        if buf.read(8) != lcmt_acrobot_y._get_packed_fingerprint():
            raise ValueError("Decode error")
        return lcmt_acrobot_y._decode_one(buf)

    @staticmethod
    def _decode_one(buf):
        self = lcmt_acrobot_y()
        self.timestamp, self.theta1, self.theta2, self.tau = struct.unpack(">qddd", buf.read(32))
        return self

    @staticmethod
    def _get_hash_recursive(parents):
        if lcmt_acrobot_y in parents: return 0
        tmphash = (0x1b1a2a81e14c5a07) & 0xffffffffffffffff
        tmphash  = (((tmphash<<1)&0xffffffffffffffff) + (tmphash>>63)) & 0xffffffffffffffff
        return tmphash
    _packed_fingerprint = None

    @staticmethod
    def _get_packed_fingerprint():
        if lcmt_acrobot_y._packed_fingerprint is None:
            lcmt_acrobot_y._packed_fingerprint = struct.pack(">Q", lcmt_acrobot_y._get_hash_recursive([]))
        return lcmt_acrobot_y._packed_fingerprint

    def get_hash(self):
        """Get the LCM hash of the struct"""
        return struct.unpack(">Q", lcmt_acrobot_y._get_packed_fingerprint())[0]




"""LCM type definitions
This file automatically generated by lcm.
DO NOT MODIFY BY HAND!!!!
"""


from io import BytesIO
import struct

import drake

class lcmt_image(object):
    """ A representation of an image. """

    __slots__ = ["header", "width", "height", "row_stride", "size", "data", "bigendian", "pixel_format", "channel_type", "compression_method"]

    __typenames__ = ["drake.lcmt_header", "int32_t", "int32_t", "int32_t", "int32_t", "byte", "boolean", "int8_t", "int8_t", "int8_t"]

    __dimensions__ = [None, None, None, None, None, ["size"], None, None, None, None]

    PIXEL_FORMAT_GRAY = 0
    """ enum for pixel_format. """
    PIXEL_FORMAT_RGB = 1
    PIXEL_FORMAT_BGR = 2
    PIXEL_FORMAT_RGBA = 3
    PIXEL_FORMAT_BGRA = 4
    PIXEL_FORMAT_DEPTH = 5
    PIXEL_FORMAT_LABEL = 6
    PIXEL_FORMAT_MASK = 7
    PIXEL_FORMAT_DISPARITY = 8
    PIXEL_FORMAT_BAYER_BGGR = 9
    PIXEL_FORMAT_BAYER_RGGB = 10
    PIXEL_FORMAT_BAYER_GBRG = 11
    PIXEL_FORMAT_BAYER_GRBG = 12
    PIXEL_FORMAT_INVALID = -1
    CHANNEL_TYPE_INT8 = 0
    """ enum for channel_type. """
    CHANNEL_TYPE_UINT8 = 1
    CHANNEL_TYPE_INT16 = 2
    CHANNEL_TYPE_UINT16 = 3
    CHANNEL_TYPE_INT32 = 4
    CHANNEL_TYPE_UINT32 = 5
    CHANNEL_TYPE_FLOAT32 = 6
    CHANNEL_TYPE_FLOAT64 = 7
    CHANNEL_TYPE_INVALID = -1
    COMPRESSION_METHOD_NOT_COMPRESSED = 0
    """ enum for compression_method. """
    COMPRESSION_METHOD_ZLIB = 1
    COMPRESSION_METHOD_JPEG = 2
    COMPRESSION_METHOD_PNG = 3
    COMPRESSION_METHOD_INVALID = -1

    def __init__(self):
        self.header = drake.lcmt_header()
        """
        The timestamp and the frame name where this image is obtained.
        LCM Type: drake.lcmt_header
        """

        self.width = 0
        """
        The image width in pixels.
        LCM Type: int32_t
        """

        self.height = 0
        """
        The image height in pixels.
        LCM Type: int32_t
        """

        self.row_stride = 0
        """
        The physical memory size per a single row in bytes.
        LCM Type: int32_t
        """

        self.size = 0
        """
        The size of `data` in bytes.
        LCM Type: int32_t
        """

        self.data = b""
        """
        The data that contains actual image.
        LCM Type: byte[size]
        """

        self.bigendian = False
        """
        The boolean to denote if the data is stored in the bigendian order.
        LCM Type: boolean
        """

        self.pixel_format = 0
        """
        The semantic meaning of pixels.
        LCM Type: int8_t
        """

        self.channel_type = 0
        """
        The data type for a channel.
        LCM Type: int8_t
        """

        self.compression_method = 0
        """
        The compression method.
        LCM Type: int8_t
        """


    def encode(self):
        buf = BytesIO()
        buf.write(lcmt_image._get_packed_fingerprint())
        self._encode_one(buf)
        return buf.getvalue()

    def _encode_one(self, buf):
        assert self.header._get_packed_fingerprint() == drake.lcmt_header._get_packed_fingerprint()
        self.header._encode_one(buf)
        buf.write(struct.pack(">iiii", self.width, self.height, self.row_stride, self.size))
        buf.write(bytearray(self.data[:self.size]))
        buf.write(struct.pack(">bbbb", self.bigendian, self.pixel_format, self.channel_type, self.compression_method))

    @staticmethod
    def decode(data: bytes):
        if hasattr(data, 'read'):
            buf = data
        else:
            buf = BytesIO(data)
        if buf.read(8) != lcmt_image._get_packed_fingerprint():
            raise ValueError("Decode error")
        return lcmt_image._decode_one(buf)

    @staticmethod
    def _decode_one(buf):
        self = lcmt_image()
        self.header = drake.lcmt_header._decode_one(buf)
        self.width, self.height, self.row_stride, self.size = struct.unpack(">iiii", buf.read(16))
        self.data = buf.read(self.size)
        self.bigendian = bool(struct.unpack('b', buf.read(1))[0])
        self.pixel_format, self.channel_type, self.compression_method = struct.unpack(">bbb", buf.read(3))
        return self

    @staticmethod
    def _get_hash_recursive(parents):
        if lcmt_image in parents: return 0
        newparents = parents + [lcmt_image]
        tmphash = (0xcc69e8044fc49594+ drake.lcmt_header._get_hash_recursive(newparents)) & 0xffffffffffffffff
        tmphash  = (((tmphash<<1)&0xffffffffffffffff) + (tmphash>>63)) & 0xffffffffffffffff
        return tmphash
    _packed_fingerprint = None

    @staticmethod
    def _get_packed_fingerprint():
        if lcmt_image._packed_fingerprint is None:
            lcmt_image._packed_fingerprint = struct.pack(">Q", lcmt_image._get_hash_recursive([]))
        return lcmt_image._packed_fingerprint

    def get_hash(self):
        """Get the LCM hash of the struct"""
        return struct.unpack(">Q", lcmt_image._get_packed_fingerprint())[0]




"""LCM type definitions
This file automatically generated by lcm.
DO NOT MODIFY BY HAND!!!!
"""


from io import BytesIO
import struct

class lcmt_point_cloud_field(object):
    """
    Describes one field (i.e., channel) within an lcmt_point_cloud.
    
    Modeled after PCL and ROS conventions:
    https://pointclouds.org/documentation/structpcl_1_1_p_c_l_point_field.html
    https://docs.ros.org/en/api/sensor_msgs/html/msg/PointField.html
    http://wiki.ros.org/pcl/Overview#Common_PointCloud2_field_names
    
    """

    __slots__ = ["name", "byte_offset", "datatype", "count"]

    __typenames__ = ["string", "int32_t", "int8_t", "int32_t"]

    __dimensions__ = [None, None, None, None]

    INT8 = 1
    """ Allowed values for datatype. """
    UINT8 = 2
    INT16 = 3
    UINT16 = 4
    INT32 = 5
    UINT32 = 6
    FLOAT32 = 7
    FLOAT64 = 8

    def __init__(self):
        self.name = ""
        """
        Field name.
        LCM Type: string
        """

        self.byte_offset = 0
        """
        Location of this field after the start of each point's data.
        LCM Type: int32_t
        """

        self.datatype = 0
        """
        Element type, per the constants shown below.
        LCM Type: int8_t
        """

        self.count = 0
        """
        Number of elements per field.
        LCM Type: int32_t
        """


    def encode(self):
        buf = BytesIO()
        buf.write(lcmt_point_cloud_field._get_packed_fingerprint())
        self._encode_one(buf)
        return buf.getvalue()

    def _encode_one(self, buf):
        __name_encoded = self.name.encode('utf-8')
        buf.write(struct.pack('>I', len(__name_encoded)+1))
        buf.write(__name_encoded)
        buf.write(b"\0")
        buf.write(struct.pack(">ibi", self.byte_offset, self.datatype, self.count))

    @staticmethod
    def decode(data: bytes):
        if hasattr(data, 'read'):
            buf = data
        else:
            buf = BytesIO(data)
        if buf.read(8) != lcmt_point_cloud_field._get_packed_fingerprint():
            raise ValueError("Decode error")
        return lcmt_point_cloud_field._decode_one(buf)

    @staticmethod
    def _decode_one(buf):
        self = lcmt_point_cloud_field()
        __name_len = struct.unpack('>I', buf.read(4))[0]
        self.name = buf.read(__name_len)[:-1].decode('utf-8', 'replace')
        self.byte_offset, self.datatype, self.count = struct.unpack(">ibi", buf.read(9))
        return self

    @staticmethod
    def _get_hash_recursive(parents):
        if lcmt_point_cloud_field in parents: return 0
        tmphash = (0xe93146ac7d998d9f) & 0xffffffffffffffff
        tmphash  = (((tmphash<<1)&0xffffffffffffffff) + (tmphash>>63)) & 0xffffffffffffffff
        return tmphash
    _packed_fingerprint = None

    @staticmethod
    def _get_packed_fingerprint():
        if lcmt_point_cloud_field._packed_fingerprint is None:
            lcmt_point_cloud_field._packed_fingerprint = struct.pack(">Q", lcmt_point_cloud_field._get_hash_recursive([]))
        return lcmt_point_cloud_field._packed_fingerprint

    def get_hash(self):
        """Get the LCM hash of the struct"""
        return struct.unpack(">Q", lcmt_point_cloud_field._get_packed_fingerprint())[0]




# noqa: shebang
"""
Provides information on the external fork of `pybind11` used by `pydrake`.
"""

repository = "pybind/pybind11"
commit = "v2.13.6"
sha256 = "e08cb87f4773da97fa7b5f035de8763abc656d87d5773e62f6da0587d1f0ec20"



# Copyright (c) 2001-2002 Enthought, Inc. 2003-2022, SciPy Developers.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided
# with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived
# from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
###############################################################################
#
# The code in this file is an extraction of the relevant parts of code from
# scipy.interpolate.interp1d for the purposes of animation in
# pydrake.systems.pyplot_visualizer.PyPlotVisualizer.
#
# This function is *NOT* intended to be used externally, it should only be used
# internally by the PyPlotVisualizer class.
import numpy as np


def _resample_interp1d(t, x, time_step):
    """
    A linear interpolation only version of scipy.interpolate.interp1d.

    Args:
        t : (N,) array_like
            A 1-D array of real values.
        x : (...,N,...) array_like
            A N-D array of real values. The length of `y` along the
            interpolation axis must be equal to the length of `x`.
        time_step: The simulation time_step, used to resample t
            evenly between min(t) and the last time with a distance of
            time_step between each unit.

    Note:
        In order to correctly interpolate values, the time sequence described
        by ``t`` must be sorted in increasing order.  In the
        rare event that the ``t`` are **not** sorted, the returned values
        **will** be sorted.  Both ``t`` and ``x`` are
        copied and sorted first, then resampled and interpolated.

    Returns:
        (t, x): The newly sampled (times, data) for plotting, their dimensions
            will be the same along the corresponding axes (rows of t and cols
            of x are equivalent).
    """
    # We may not assume that the provided sample times are sorted in general.
    # We may assume, however, that t and x have equitable dimensions.
    # The number of t rows and x cols are the same, and we will
    # process x on the cols (axis=1).
    axis = 1
    x_copy = np.array(x, copy=True)
    t_copy = np.array(t, copy=True)
    sort_indices = np.argsort(t_copy, kind="mergesort")
    t_copy = t_copy[sort_indices]
    x_copy = np.take(x_copy, sort_indices, axis=axis)

    # Now that we are sorted, resample the times and data.
    t_resample = np.arange(t_copy[0], t_copy[-1], time_step)
    t_new_indices = np.searchsorted(t_copy, t_resample)
    t_new_indices = t_new_indices.clip(1, len(t_copy) - 1).astype(int)
    lo_indices = t_new_indices - 1
    hi_indices = t_new_indices

    # Resample times.
    t_lo = t_copy[lo_indices]
    t_hi = t_copy[hi_indices]

    # Resample data.
    x_axis = axis % x_copy.ndim
    xi = np.moveaxis(np.asarray(x_copy), x_axis, 0)
    xi = xi.reshape((xi.shape[0], -1))
    xi_lo = xi[lo_indices]
    xi_hi = xi[hi_indices]

    # Perform the linear interpolation.
    slope = (xi_hi - xi_lo) / (t_hi - t_lo)[:, None]
    xi_new = slope * (t_resample - t_lo)[:, None] + xi_lo
    x_extra_shape = x_copy.shape[:x_axis] + x_copy.shape[x_axis + 1 :]
    x_final = xi_new.reshape(t_resample.shape + x_extra_shape)
    if x_axis != 0 and t_resample.shape != ():
        nx = len(t_resample.shape)
        ny = len(x_extra_shape)
        s = (
            list(range(nx, nx + x_axis))
            + list(range(nx))
            + list(range(nx + x_axis, nx + ny))
        )
        x_final = x_final.transpose(s)

    return t_resample, x_final
